package com.htfp.weather.griddata.common;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.htfp.weather.download.GfsDataConfig;
import com.htfp.weather.utils.JSONUtils;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import ucar.ma2.DataType;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.NoSuchFileException;
import java.util.List;

@Data @Component @Slf4j
@JsonIgnoreProperties(ignoreUnknown = true)
public class TableConfig {
    @Resource
    @JsonIgnore
    GfsDataConfig dataConfig;

    @Value("duration")
    public String dataTableName;
    public String metaTableName;
    public String dataIndexName;
    public String metaIndexName;

    /**
     * 数据属性：ID、本地数据文件名、目标变量名、变量维度大小、变量类型
     */
    public String dataDir;
    public List<String> variableList;
    public int lonSize;
    public int latSize;
    public int levSize;
    public int timeToLive;
    
    @JsonIgnore
    public int timeSizeMax = 99;
    @JsonIgnore
    public double[] lonList;
    @JsonIgnore
    public double[] latList;
    @JsonIgnore
    public int[] pressureList;
    @JsonIgnore
    public int[] heightList;
    // @JsonIgnore
    // public List<Integer> forecastTimeList; // 预报时效列表
    @JsonIgnore
    public DataType dataType = DataType.FLOAT;

    // note 修改作用范围在classes目录下，resource文件下不改变
    @JsonIgnore
    // private final String configPath = Objects.requireNonNull(this.getClass().getClassLoader().getResource("config/tableConf.json")).getPath();
    private final String configPath = System.getProperty("user.dir") +"/tableConf.json";
    @PostConstruct
    private void initTableConfig() {
        readConfig();
        initLonList();
        initLatList();
        initLevList();
    }

    public void readConfig() {
        // String pathSeparator = System.getProperty("file.separator");
        try (InputStream f = new FileInputStream(configPath)){
            String jsonStr = IOUtils.toString(f, StandardCharsets.UTF_8);
            TableConfig tableConfig = JSONUtils.json2pojo(jsonStr, TableConfig.class);
            this.dataTableName = tableConfig.getDataTableName();
            this.metaTableName = tableConfig.getMetaTableName();
            this.dataIndexName = tableConfig.getDataIndexName();
            this.metaIndexName = tableConfig.getMetaIndexName();
            this.variableList = tableConfig.getVariableList();
            this.dataDir = tableConfig.getDataDir();
            this.lonSize = tableConfig.getLonSize();
            this.latSize = tableConfig.getLatSize();
            this.levSize = dataConfig.getPressureLevels().length;
            this.timeToLive = tableConfig.getTimeToLive();
            log.info("[config] 读取 DataTable 配置 : {}", configPath);
        } catch (NoSuchFileException e) {
            log.error("[config] 配置文件{}不存在", configPath, e);
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

    public void writeConfig(TableConfig updateConfig) throws IOException {
        writeConfig(updateConfig, configPath);
    }
    public void writeConfig(TableConfig updateConfig, String filePath) throws IOException {
        this.dataTableName = updateConfig.getDataTableName();
        this.metaTableName = updateConfig.getMetaTableName();
        this.dataIndexName = updateConfig.getDataIndexName();
        this.metaIndexName = updateConfig.getMetaIndexName();
        this.variableList = updateConfig.getVariableList();
        this.dataDir = updateConfig.getDataDir();
        this.lonSize = updateConfig.getLonSize();
        this.latSize = updateConfig.getLatSize();
        this.levSize = updateConfig.getLevSize();
        // // FIXME 2024/5/10: 维度信息取决于数据下载，目前需要两个配置一起更新
        this.timeToLive = updateConfig.getTimeToLive();
        JSONUtils.pojo2jsonFile(this, filePath);
        log.info("配置文件 {} 更新为: {}", filePath, updateConfig);
    }

    public void valid() {
    }


    private void initLonList() {
        double lonStart = dataConfig.getMinLon();
        double res = dataConfig.getResolution();
        lonList = new double[lonSize];
        for (int i = 0; i < lonSize; i++) {
            lonList[i] = lonStart + i * res;
        }
    }

    private void initLatList() {
        double latStart = dataConfig.getMinLat();
        double res = dataConfig.getResolution();
        latList = new double[latSize];
        for (int i = 0; i < latSize; i++) {
            latList[i] = latStart + i * res;
        }
    }

    private void initLevList() {
        pressureList = dataConfig.getPressureLevels();
        heightList = dataConfig.getHeightLevels();
        // this.levSize = Math.max(pressureList.length, heightList.length);
    }
}
