package com.htfp.weather.web.controller;

import com.htfp.weather.download.FileInfo;
import com.htfp.weather.download.GfsDataConfig;
import com.htfp.weather.download.GfsDownloader;
import com.htfp.weather.griddata.common.TableConfig;
import com.htfp.weather.griddata.operation.GfsDataImport;
import com.htfp.weather.griddata.operation.GfsDataImport.ImportResult;
import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.web.exception.AppException;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.request.GfsConfigUpdate;
import com.htfp.weather.web.pojo.request.TableConfigUpdate;
import com.htfp.weather.web.pojo.response.Result;
import com.htfp.weather.web.service.FileService;
import org.apache.commons.lang3.StringUtils;
import org.springframework.util.CollectionUtils;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import java.io.File;
import java.io.IOException;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/4/21 15:56
 * @Description : 配置文件管理
 */
@CrossOrigin
@RestController
@RequestMapping("/htfp/weather/config")
public class ConfigController {
    @Resource
    TableConfig tableConfig;
    @Resource
    GfsDataConfig gfsDataConfig;
    @Resource
    GfsDownloader gfsDownloader;
    @Resource
    GfsDataImport gfsDataImport;
    @Resource
    FileService fileService;
    private final String SECRET = "htfpweather";

    private void validSecret(String secret) {
        if (StringUtils.isEmpty(secret) || !SECRET.equals(secret)) {
            throw new AppException(ErrorCode.SECRET_ERROR);
        }
    }

    /**
     * 查询tablestore数据库配置
     */
    @RequestMapping("/queryDatabaseConfig")
    public Result queryDatabaseConfig() {
        return Result.success(tableConfig);
    }

    /**
     * 更新tablestore数据库配置
     */
    @PostMapping("/updateDatabaseConfig")
    public Result updateDatabaseConfig(@RequestBody TableConfigUpdate request) {
        String secret = request.getSecret();
        TableConfig updateConfig = request.getConfig();
        validSecret(secret);
        if (updateConfig == null) {
            throw new AppException(ErrorCode.CONFIG_ERROR, "配置文件不能为空");
        }
        updateConfig.valid();
        try {
            tableConfig.writeConfig(updateConfig);
        } catch (IOException e) {
            throw new AppException(ErrorCode.CONFIG_ERROR, e);
        }
        return Result.success();
    }

    /**
     * 查询GFS下载配置
     */
    @RequestMapping("/queryDataSourceConfig")
    public Result queryDataSourceConfig() {
        return Result.success(gfsDataConfig);
    }

    /**
     * 更新GFS下载配置
     */
    @PostMapping("/updateDataSourceConfig")
    public Result updateDataSourceConfig(@Validated @RequestBody GfsConfigUpdate request) {
        String secret = request.getSecret();
        GfsDataConfig updateConfig = request.getConfig();
        validSecret(secret);
        if (updateConfig == null) {
            throw new AppException(ErrorCode.CONFIG_ERROR, "配置文件不能为空");
        }
        updateConfig.valid();
        try {
            gfsDataConfig.writeConfig(updateConfig);
        } catch (IOException e) {
            throw new AppException(ErrorCode.CONFIG_ERROR, e);
        }
        return Result.success();
    }

    /**
     * 下载当前时刻未来24小时的数据文件
     */
    @PostMapping("/downloadAllFiles")
    public Result downloadAllFiles(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        // List<FileInfo> allCompleted = false;
        try {
            gfsDownloader.iniTimeSetting();
            List<FileInfo> fileInfoList = gfsDownloader.getFilesInfo();
            List<FileInfo> finishedList = gfsDownloader.downloadAll(fileInfoList);
            List<FileInfo> successList = finishedList.stream().filter(FileInfo::isDownloadSuccess).collect(Collectors.toList());
            List<FileInfo> failedList = finishedList.stream().filter(fileInfo -> !fileInfo.isDownloadSuccess()).collect(Collectors.toList());
            Map<String, List<FileInfo>> data = new HashMap<>();
            data.put("successList", successList);
            data.put("failedList", failedList);
            if (CollectionUtils.isEmpty(failedList)) {
                return Result.success(data);
            } else {
                return new Result(false, ErrorCode.DOWNLOAD_ERROR.getCode(), ErrorCode.DOWNLOAD_ERROR.getMsg() + ":下载未全部完成", data);
            }
        } catch (AppException e) {
            return Result.error(ErrorCode.DOWNLOAD_START_ERROR, e.getMessage());
        }
    }

    /**
     * 下载指定时刻文件
     */
    @PostMapping("/downloadSingleFile")
    public Result downloadSingleFile(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        OffsetDateTime offsetDateTime = OffsetDateTime.parse(params.get("time"), Constant.API_TIME_FORMATTER);
        String targetUtcStr = DateTimeUtils.getUTCDateTime(offsetDateTime).format(DateTimeFormatter.ofPattern(Constant.UTC_TIME_STRING));
        try {
            gfsDownloader.iniTimeSetting();
            List<FileInfo> fileInfoList = gfsDownloader.getFilesInfo(); // TODO 2024/6/12: 可以优化，不用生成24小时所有的文件信息
            for (FileInfo fileInfo : fileInfoList) {
                if (fileInfo.getForecastUTCTimeStr().equals(targetUtcStr)) {
                    FileInfo downloadResult = gfsDownloader.download(fileInfo);
                    if (downloadResult.isDownloadSuccess()) {
                        return Result.success(downloadResult);
                    } else {
                        return Result.error(ErrorCode.DOWNLOAD_ERROR);
                    }
                }
            }
            return Result.error(ErrorCode.QUERY_TIME_ERROR, ": 请选择当前时间之后24小时范围内的时刻");
        } catch (AppException e) {
            return Result.error(ErrorCode.DOWNLOAD_START_ERROR, e.getMessage());
        }
    }

    /**
     * 根据起报时间将数据导入数据库
     */
    @PostMapping("/importToTablestoreByReferenceTime")
    public Result importData(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        OffsetDateTime time = OffsetDateTime.parse(params.get("time"), Constant.API_TIME_FORMATTER);
        try {
            List<ImportResult> finishedList = gfsDataImport.importData(time);
            List<ImportResult> successList = finishedList.stream().filter(result -> result.isSuccess()).collect(Collectors.toList());
            List<ImportResult> failedList = finishedList.stream().filter(result -> !result.isSuccess()).collect(Collectors.toList());
            Map<String, List<ImportResult>> data = new HashMap<>();
            data.put("successList", successList);
            data.put("failedList", failedList);
            if (CollectionUtils.isEmpty(failedList)) {
                return Result.success(data);
            } else {
                return new Result(false, ErrorCode.TABLESTORE_IMPORT_ERROR.getCode(), ErrorCode.TABLESTORE_IMPORT_ERROR.getMsg() + ": 数据导入未全部完成", data);
            }
        } catch (AppException e) {
            return Result.error(e.getErrorCode());
        } catch (Exception e) {
            return Result.error(e);
        }
    }

    @PostMapping("queryAllDataDir")
    public Result queryAllDataDir(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        File dataRoot = new File(gfsDataConfig.getSaveRoot());
        List<String> subDirList = fileService.getSubDirList(dataRoot);
        return Result.success(subDirList);
    }
    /**
     * 查询指定起报时间文件夹中的所有文件名
     */
    @PostMapping("/queryAllFileByReferenceTime")
    public Result queryAllFileByReferenceTime(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        OffsetDateTime time = OffsetDateTime.parse(params.get("time"), Constant.API_TIME_FORMATTER);
        String timeStr = DateTimeUtils.getUTCDateTime(time).format(DateTimeFormatter.ofPattern(Constant.DATA_FOLDER_STRING));
        File dataDir = new File(gfsDataConfig.getSaveRoot(), timeStr);

        // 查找目录下的文件
        List<com.htfp.weather.web.pojo.response.FileInfo> fileList = fileService.getFileList(dataDir);
        return Result.success(fileList);
    }

    /**
     * 删除指定起报时间对应的文件夹
     */
    @PostMapping("/deleteDirByReferenceTime")
    public Result deleteDirByReferenceTime(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        OffsetDateTime time = OffsetDateTime.parse(params.get("time"), Constant.API_TIME_FORMATTER);
        String timeStr = DateTimeUtils.getUTCDateTime(time).format(DateTimeFormatter.ofPattern(Constant.DATA_FOLDER_STRING));
        File dataDir = new File(gfsDataConfig.getSaveRoot(), timeStr);
        if (gfsDataImport.isImporting()) {
            return Result.error(ErrorCode.CAN_NOT_DELETE_WHEN_IMPORT_OR_DOWNLOAD);
        }
        try {
            fileService.deleteDir(dataDir);
            return Result.success();
        } catch (IOException e) {
            return Result.error(e);
        }
    }

    /**
     * 删除指定文件
     */
    @PostMapping("/deleteTargetFile")
    public Result deleteTargetFile(@RequestBody Map<String, String> params) {
        String secret = params.get("secret");
        validSecret(secret);
        String filePath = params.get("filePath");
        File file = new File(filePath);
        if (gfsDataImport.isImporting()) {
            return Result.error(ErrorCode.CAN_NOT_DELETE_WHEN_IMPORT_OR_DOWNLOAD);
        }
        try {
            fileService.deleteFile(file);
            return Result.success();
        } catch (IOException e) {
            return Result.error(e);
        }
    }
}
