package com.htfp.weather.web.service;

import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Grid4D;
import com.htfp.weather.griddata.common.GfsVariableHeightEnum;
import com.htfp.weather.griddata.common.GfsVariableIsobaricEnum;
import com.htfp.weather.griddata.common.TableConfig;
import com.htfp.weather.griddata.operation.GfsDataFetcher;
import com.htfp.weather.griddata.utils.ValueRange;
import com.htfp.weather.info.Constant;
import com.htfp.weather.info.GfsLevelsEnum;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.web.exception.AppException;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.response.NowWeatherStatus;
import com.htfp.weather.web.pojo.response.PlaneResponse;
import com.htfp.weather.web.pojo.response.ProfileResponse;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;
import ucar.ma2.Array;
import ucar.ma2.Index4D;

import javax.annotation.Resource;
import java.lang.reflect.Method;
import java.time.Duration;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;

/**
 * @Author : shiyi
 * @Date : 2024/2/2 15:47
 * @Description : 获取tablestore格点数据
 */
@Service("tablestore-gfs")
@Slf4j
public class GfsDataServiceImpl implements IDataService {
    @Resource
    GfsDataFetcher gfsDataFetcher;
    @Resource
    TableConfig tableConfig;



    /**
     * 获取当前时刻的天气状况
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link NowWeatherStatus}
     */
    public NowWeatherStatus getNowWeather(int level, double latitude, double longitude) {
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(nowTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(Collections.singletonList(latitude));
        ValueRange<Double> lonRange = new ValueRange<>(Collections.singletonList(longitude));
        List<String> variableNames = getVariableNamesInDatabase(level);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames,timeRange, levRange, latRange, lonRange
            );
            final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
            // 每个变量都是0维数组
            NowWeatherStatus nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, level, latitude, longitude, levelFlag);
            nowWeatherStatus.setLatitude(latitude);
            nowWeatherStatus.setLongitude(longitude);
            return nowWeatherStatus;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获得指定点序列上的当前气象信息
     * @param latitude
     * @param longitude
     * @param level
     * @return
     */
    public List<NowWeatherStatus> getNowWeatherByMultiPoint(int level, List<Double> latitude, List<Double> longitude) {
        if (latitude.size() != longitude.size()) {
            throw new AppException(ErrorCode.INDEX_SIZE_ERROR, ": 经纬度数组大小不一致");
        }
        int pointSize = latitude.size();
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(nowTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(latitude);
        ValueRange<Double> lonRange = new ValueRange<>(longitude);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames,timeRange, levRange, latRange, lonRange
            );
            List<NowWeatherStatus> nowWeatherStatusList = new ArrayList<>();

            for (int i = 0; i < pointSize; i++) {
                NowWeatherStatus nowWeatherStatus;
                // NOTE 2024/6/11: build中传入的索引和数据库的查询索引意义不同
                final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
                nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, level, latitude.get(i), longitude.get(i), levelFlag);
                nowWeatherStatusList.add(nowWeatherStatus);
            }
            return nowWeatherStatusList;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取未来24小时的气象时间序列数据
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link TimeSeriesDataset}
     */
    public TimeSeriesDataset getForecastSeries(int level, double latitude, double longitude) {
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Arrays.asList(nowTime, nowTime.plusHours(24)));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(Collections.singletonList(latitude));
        ValueRange<Double> lonRange = new ValueRange<>(Collections.singletonList(longitude));

        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            );
            final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
            return buildTimeSeriesDatasetInTargetPoint(gridDataSet, level, latitude, longitude, levelFlag);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获得指定点序列上的未来24小时气象预报
     * @param latitudeList
     * @param longitudeList
     * @param level
     * @return
     */
    public List<TimeSeriesDataset> getForecastSeriesByMultiPoint(List<Double> latitudeList, List<Double> longitudeList, int level) {
        if (latitudeList.size() != longitudeList.size()) {
            throw new AppException(ErrorCode.INDEX_SIZE_ERROR, ": 经纬度数组大小不一致");
        }
        int pointSize = latitudeList.size();
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Arrays.asList(nowTime, nowTime.plusHours(24)));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(latitudeList);
        ValueRange<Double> lonRange = new ValueRange<>(longitudeList);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            );
            List<TimeSeriesDataset> timeSeriesDatasetList = new ArrayList<>();
            for (int i = 0; i < pointSize; i++) {
                final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
                TimeSeriesDataset timeSeriesDataset = buildTimeSeriesDatasetInTargetPoint(gridDataSet, level, latitudeList.get(i), longitudeList.get(i), levelFlag);
                timeSeriesDatasetList.add(timeSeriesDataset);
            }
            return timeSeriesDatasetList;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取变量随气压变化的廓线
     *
     * @param targetTime   目标时间 UTC
     * @param variableName 数据库中的变量名
     * @param latitude     纬度
     * @param longitude    经度
     * @return {@link ProfileResponse}
     */
    public ProfileResponse getProfileByVariableAndPressure(String variableName, OffsetDateTime targetTime, double latitude, double longitude) {
        String targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        if (targetVariable == null) {
            throw new AppException(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        try {
            int[] pressureLevels = tableConfig.getPressureList();
            Array array = gfsDataFetcher.getProfileByVariableAndPressure(targetVariable, targetTime, latitude, longitude);
            float[] values = (float[]) array.copyTo1DJavaArray();
            ProfileResponse profileResponse = new ProfileResponse(pressureLevels, values);
            profileResponse.setLatitude(latitude);
            profileResponse.setLongitude(longitude);
            return profileResponse;
        } catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }

    }

    /**
     * 获取指定变量、时间、高度的二维数组数据
     *
     * @param targetTime   目标时间 UTC
     * @param variableName api中的变量名
     * @param level        高度
     * @param minLat       最小纬度
     * @param maxLat       最大纬度
     * @param minLon       最小经度
     * @param maxLon       最大经度
     * @return @{@link PlaneResponse}
     */
    public PlaneResponse getPlane(OffsetDateTime targetTime, String variableName, int level, double minLat, double maxLat, double minLon, double maxLon) {
        String targetVariable;
        if (GfsLevelsEnum.SURFACE.equals(GfsLevelsEnum.getByCode(level))) {
            targetVariable = GfsVariableHeightEnum.getGfsVariableName(variableName);
        } else {
            targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        }

        if (targetVariable == null) {
            throw new AppException(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }

        List<String> variableNames = Collections.singletonList(targetVariable);
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(targetTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(minLat, maxLat);
        ValueRange<Double> lonRange = new ValueRange<>(minLon, maxLon);

        try {
            Array array = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            ).getVariable(targetVariable).toArray();
            return buildPlane(array, minLat, maxLat, minLon, maxLon);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }


    private List<String> getVariableNamesInDatabase(int level) {
        List<String> variableNames = new ArrayList<>();
        if (GfsLevelsEnum.SURFACE.equals(GfsLevelsEnum.getByCode(level))) {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableHeightEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        } else {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableIsobaricEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        }
        return variableNames;
    }

    private PlaneResponse buildPlane(Array array, double minLat, double maxLat, double minLon, double maxLon) {
        int iMinLat = gfsDataFetcher.getLatitudeIndex(minLat);
        int iMaxLat = gfsDataFetcher.getLatitudeIndex(maxLat);
        int iMinLon = gfsDataFetcher.getLongitudeIndex(minLon);
        int iMaxLon = gfsDataFetcher.getLongitudeIndex(maxLon);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;
        PlaneResponse planeResponse = new PlaneResponse();
        float[][] values = (float[][]) array.reduce().copyToNDJavaArray();
        planeResponse.setLatList(Arrays.copyOfRange(tableConfig.getLatList(), iMinLat, iMinLat + latLength));
        planeResponse.setLonList(Arrays.copyOfRange(tableConfig.getLonList(), iMinLon, iMinLon + lonLength));
        planeResponse.setValues(values);
        return planeResponse;
    }


    /**
     * 构建指定位置的预报数据
     * @param gridDataSet
     * @param level
     * @param latitude
     * @param longitude
     * @param levelFlag-高度层次标记，地面或高空
     * @return
     */

    private TimeSeriesDataset buildTimeSeriesDatasetInTargetPoint(GridDataSet gridDataSet, int level, double latitude, double longitude,GfsLevelsEnum levelFlag) {
        TimeSeriesDataset result = new TimeSeriesDataset();
        result.setLatitude(latitude);
        result.setLongitude(longitude);
        // 获取全局索引
        int levIndexGlobal = gfsDataFetcher.getLevelIndex(level);
        int latIndexGlobal = gfsDataFetcher.getLatitudeIndex(latitude);
        int lonIndexGlobal = gfsDataFetcher.getLongitudeIndex(longitude);

        // 利用反射构建数据集
        Class<? extends TimeSeriesDataset> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            String variableNameInFile = entry.getKey();
            String variableNameInApi;
            if (GfsLevelsEnum.SURFACE.equals(levelFlag)) {
                variableNameInApi = GfsVariableHeightEnum.getVariableNameInApi(variableNameInFile);
            } else {
                variableNameInApi = GfsVariableIsobaricEnum.getVariableNameInApi(variableNameInFile);
            }
            if (variableNameInApi == null) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE);
            }
            Grid4D grid4D = entry.getValue();
            int[] origin = grid4D.getOrigin(); // grid4D的起始坐标全局索引
            int[] shape = grid4D.getShape();
            int tSize = shape[0];
            try {
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                // TODO 2024/6/20: 这里的toArray操作在数据量较大的时候会比较吃内存和cpu
                float[] storage = (float[]) grid4D.toArray().section(
                        new int[]{0, levIndexGlobal-origin[1], latIndexGlobal-origin[2], lonIndexGlobal-origin[3]}, new int[]{tSize, 1, 1, 1}
                ).copyTo1DJavaArray();
                method.invoke(result, storage);
            } catch (ReflectiveOperationException e) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppException(ErrorCode.RESPONSE_DATA_BUILD_ERROR, e);
            }

            // 设置时间
            if (result.getTime() == null) {
                List<String> timeList = getResultTimeStringList(gridDataSet, tSize, origin[0]);
                result.setTime(timeList);
            }
        }
        return result;
    }


    private List<String> getResultTimeStringList(GridDataSet gridDataSet, int timeSize, int iTimeStart) {
        GridDataSetMeta meta = gridDataSet.getMeta();
        List<String> forecastHours = meta.getForecastHours();
        long milli = (long) meta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        List<String> timeList = new ArrayList<>();
        for (int i = 0; i < timeSize; i++) {
            int hour = Integer.parseInt(forecastHours.get(i + iTimeStart));
            String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                    .format(Constant.API_TIME_FORMATTER);
            timeList.add(timeStr);
        }
        return timeList;
    }

    /**
     * 构建指定位置的当前天气数据
     * @param gridDataSet
     * @param level
     * @param latitude
     * @param longitude
     * @param levelFlag-高度层次标记，地面或高空
     * @return
     */
    private NowWeatherStatus buildNowWeatherInTargetPoint(GridDataSet gridDataSet, int level, double latitude, double longitude,  GfsLevelsEnum levelFlag) {
        String timeStr = OffsetDateTime.now().withMinute(0).withNano(0).format(Constant.API_TIME_FORMATTER);
        NowWeatherStatus result = new NowWeatherStatus();
        result.setTime(timeStr);
        result.setLatitude(latitude);
        result.setLongitude(longitude);
        // 获取全局索引
        int levIndexGlobal = gfsDataFetcher.getLevelIndex(level);
        int latIndexGlobal = gfsDataFetcher.getLatitudeIndex(latitude);
        int lonIndexGlobal = gfsDataFetcher.getLongitudeIndex(longitude);

        // 利用反射构建数据集
        Class<? extends NowWeatherStatus> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            String variableNameInFile = entry.getKey();
            String variableNameInApi;
            if (GfsLevelsEnum.SURFACE.equals(levelFlag)) {
                variableNameInApi = GfsVariableHeightEnum.getVariableNameInApi(variableNameInFile);
            } else {
                variableNameInApi = GfsVariableIsobaricEnum.getVariableNameInApi(variableNameInFile);
            }
            if (variableNameInApi == null) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE);
            }
            Grid4D grid4D = entry.getValue(); Index4D index = new Index4D(grid4D.getShape());
            int[] origin = grid4D.getOrigin(); // grid4D的起始坐标全局索引
            try {
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                // TODO 2024/6/19: 避免toArray操作，减少内存复制
                method.invoke(result, grid4D.toArray().getFloat(
                        index.set(0, levIndexGlobal-origin[1], latIndexGlobal-origin[2], lonIndexGlobal-origin[3])
                ));
            } catch (ReflectiveOperationException e) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppException(ErrorCode.RESPONSE_DATA_BUILD_ERROR);
            }
        }
        return result;
    }


    private Map<String, List> getFutureTimeIndexList(GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        OffsetDateTime currentTime = DateTimeUtils.getUTCDateTime(OffsetDateTime.now());
        if (currentTime.isBefore(refTime)) {
            throw new AppException(ErrorCode.QUERY_TIME_ERROR);
        }
        int startHour = getForecastHourFromTargetTime(currentTime, refTime);
        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        List<Integer> iTimeList = new ArrayList<>();
        List<String> timeList = new ArrayList<>();

        for (int i = 0; i < forecastHours.size(); i++) {
            int hour = Integer.parseInt(forecastHours.get(i));
            if (hour >= startHour) {
                iTimeList.add(i);
                String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                        .format(Constant.API_TIME_FORMATTER);
                timeList.add(timeStr);
            }
            if (iTimeList.size() >= 24) {
                break;
            }
        }
        if (CollectionUtils.isEmpty(iTimeList) || CollectionUtils.isEmpty(timeList)) {
            throw new AppException(ErrorCode.QUERY_TIME_ERROR);
        }
        Map<String, List> map = new HashMap<>(2);
        map.put("iTimeList", iTimeList);
        map.put("timeList", timeList);
        return map;
    }

    private int getForecastHourFromTargetTime(OffsetDateTime targetTime, OffsetDateTime refTime) {
        return (int) Duration.between(refTime, targetTime).toHours();
    }

}
